@@ -2,7 +2,7 @@ from __future__ import annotations
 
 import operator
 from functools import wraps
-from typing import Hashable, FrozenSet, Callable, Optional
+from typing import Hashable, FrozenSet, Callable, Optional, Sequence
 
 import attr
 import funcy as fn
@@ -16,13 +16,13 @@ Alphabet = FrozenSet[Letter]
 def boolean_only(method):
     @wraps(method)
     def wrapped(self, *args, **kwargs):
-        if self.outputs != {True, False}:
+        if not (self.outputs <= {True, False}):
             raise ValueError(f'{method} only defined for Boolean output DFAs.')
         return method(self, *args, **kwargs)
     return wrapped
 
 
-@attr.frozen
+@attr.frozen(auto_detect=True)
 class DFA:
     start: State
     _label: Callable[[State], Letter] = attr.ib(
@@ -35,23 +35,44 @@ class DFA:
         converter=lambda x: x if x is None else frozenset(x), default=None
     )
     outputs: Alphabet = attr.ib(converter=frozenset, default={True, False})
-    _states: Optional[FrozenSet[State]] = None
+    _states: Optional[Sequence[State]] = None
+    _hash: Optional[int] = None
 
     def __repr__(self) -> int:
         from dfa.utils import dfa2dict
         import pprint
 
         if self.inputs is not None:
-            return pprint.pformat(dfa2dict(self))
+            return 'DFA' + pprint.pformat(dfa2dict(self))
         else:
             start, inputs, outputs = self.start, self.inputs, self.outputs
             return f'DFA({start=},{inputs=},{outputs=})'
 
+    def normalize(self) -> DFA:
+        """Normalizes the state indexing and memoizes transitions/labels."""
+        from dfa.utils import dfa2dict
+        from dfa.utils import dict2dfa
+        return dict2dfa(*dfa2dict(self, reindex=True))
+
     def __hash__(self) -> int:
-        return hash(repr(self))
+        if self._hash is None:
+            _hash = hash(repr(self.normalize()))
+            object.__setattr__(self, "_hash", _hash)  # Cache hash.
+        return self._hash
 
     def __eq__(self, other: DFA) -> bool:
-        return isinstance(other, DFA) and (repr(self) == repr(other))
+        from dfa.utils import find_equiv_counterexample as test_equiv
+        from dfa.utils import dfa2dict
+
+        if not isinstance(other, DFA):
+            return False
+
+        bool_ = {True, False}
+        if (self.outputs <= bool_) and (other.outputs <= bool_):
+            return test_equiv(self, other) is None
+        else:
+            return dfa2dict(self, reindex=True) \
+                    == dfa2dict(other, reindex=True)
 
     def run(self, *, start=None, label=False):
         """Co-routine interface for simulating runs of the automaton.
@@ -90,26 +111,29 @@ class DFA:
         return tuple(map(self._label, self.trace(word, start=start)))[:-1]
 
     def states(self):
-        if self._states is not None:
-            return self._states
+        if self._states is None:
+            assert self.inputs is not None, "Need to specify inputs field!"
 
-        assert self.inputs is not None, "Need to specify inputs field for DFA!"
+            # Make search deterministic.
+            try:
+                inputs = sorted(self.inputs)  # Try to respect inherent order.
+            except TypeError:
+                inputs = sorted(self.inputs, key=id)  # Fall by to object ids.
 
-        visited = set()
-        stack = [self.start]
-        while stack:
-            curr = stack.pop()
-            if curr in visited:
-                continue
-            else:
-                visited.add(curr)
+            visited, order = set(), []
+            stack = [self.start]
+            while stack:
+                curr = stack.pop()
+                if curr in visited:
+                    continue
+                else:
+                    order.append(curr)
+                    visited.add(curr)
 
-            successors = [self._transition(curr, a) for a in self.inputs]
-            stack.extend(successors)
-
-        # Cache _states. Note: Safely violates frozen.
-        object.__setattr__(self, "_states", visited)
-        return visited
+                successors = [self._transition(curr, a) for a in inputs]
+                stack.extend(successors)
+            object.__setattr__(self, "_states", tuple(order))  # Cache states.
+        return frozenset(self._states)
 
     @boolean_only
     def __invert__(self):
